/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#include <netdb.h>
#include <string>
#include <dirent.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "globus_config.h"
#include "gssapi.h"
#include "globus_gss_assist.h"

#include <stdio.h>
#include <openssl/x509.h>
#include <openssl/asn1.h>

#include "newca.h"
#include "data.h"

#include "Client.h"

#include "voms_api.h"

#include <iostream>
#include <iomanip>

#include <openssl/pem.h>
#include <openssl/asn1.h>
#include <openssl/evp.h>

#include <signal.h>
#include "sign.h"
#include "api_util.h"

#include <sys/wait.h>

static bool
parse(std::string &is, const std::string s, int &val)
{
  int len = s.length();
  std::string v = is.substr(0, len);

  if (v == s) {
    val = 0;

    while(isdigit(is[len])) {
      val *= 10;
      val += is[len++]- '0';
    }
    if (is[len++] == '\n') {
      is.erase(0, len);
      return true;
    }
  }

  return false;
}

static bool  
parse(std::string &is, const std::string s, std::string &val)
{
  int len = s.length();
  std::string v = is.substr(0, len);

  if (v == s) {
    std::string::size_type pos = is.find('\n');
    if (pos != std::string::npos) {
      val = std::string(is.substr(len, pos-len));
      is.erase(0,pos+1);
      return true;
    }
  }

  return false;
}

static bool  
parse(std::string &is, const std::string s)
{
  int len = s.length();
  std::string v = is.substr(0, len);

  if (v == s) {
    is.erase(0, len);
    return true;
  }

  return false;
}


/*
 * Function:
 *   net_connect(hostname, port, contact, vo, command, commandlen, buffer, buflen)
 *
 * Description:
 *   This function connects to the VOMS server, submits a command, gets the
 *   results, verifies them and finally returns them.
 *
 * Parameters:
 *   'hostname'   - This is the hostname of the server to contact.
 *   'port'       - This is the port on which the server is listening.
 *   'contact'    - This is the subject of the server's certificate.
 *   'vo'         - The VO to whom the VOMS server belongs.
 *   'command'    - This is the command to be executed.
 *   'commandlen' - This is the command's length
 *   'buffer'     - OUTPUT PARAMETER: This buffer contains the result of
 *                  the command sent to the server.
 *   'buflen'     - OUTPUT PARAMETER: This is the length of the 'buffer'
 *                  parameter.
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     Non zero. Furthermore, the output parameters are properly initialized.
 *
 * NOTE: The memory pointed by the 'buffer' parameter may very well be
 * overwritten by consecutive calls to this function.
 */
bool
contact(const std::string &hostname, int port, const std::string &contact,
	const std::string &command, std::string &buffer, std::string &subject, std::string &ca, verror_type &error)
{
  bool res = false;
  std::string temp;

  GSISocketClient sock(hostname, port, globus(0));

  sock.RedirectGSIOutput(NULL);
  sock.ServerContact(contact);
  sock.SetFlags(GSS_C_MUTUAL_FLAG | GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG);

  if (!sock.Open()) {
    error = VERR_NOSOCKET;
    return false;
  }
#ifdef DEBUG
  cout << "**********" << endl
       << "OWN  SUBJ:" << sock.own_subject << endl
       << "OWN  CA  :" << sock.own_ca << endl
       << "PEER SUBJ:" << sock.peer_subject << endl
       << "PEER CA  :" << sock.peer_ca << endl << flush;
#endif          

  //  GSISocketAgent *sa = sock.Agent();

  if (sock.own_subject.empty()) {
    error = VERR_NOIDENT;
    goto err;
  }

  subject = sock.own_subject;
  ca      = sock.own_ca;

  if (!sock.Send(command) || !sock.Receive(buffer)) {
    error = VERR_COMM;
    goto err;
  }

  res = true;
 err:
  sock.Close();
  
  return res;
}   



bool retrieve(X509 *cert, STACK_OF(X509) *chain, recurse_type how, std::string &buffer, std::string &vo, std::string &file, std::string &subject, std::string &ca, verror_type &error)
{
  int size = 0, res;
  int len;
  bool found = false;

  if (!cert || (!chain && (how != RECURSE_NONE))) {
    error = VERR_PARAM;
    return false;
  }

  /*
   * check credential and get the globus name
   */
  int chain_length;
  X509_EXTENSION *ext;
  X509 *datacert=NULL;
  int index = 0;
  int nid = 0, nidf = 0, nidv = 0;
  ASN1_OCTET_STRING *obj;
  int position = 0;
  char buf[1000];

  error = VERR_IDCHECK;

  ca = subject = "";

  chain_length = sk_X509_num(chain);
  if (chain == NULL) {
    if (X509_NAME_oneline(X509_get_subject_name(cert), buf, 1000)) {
      subject = std::string(buf);
    }
    if (X509_NAME_oneline(X509_get_issuer_name(cert), buf, 1000)) {
      ca = std::string(buf);
    }
  }
  else {
  /* Determine id data */
    datacert = sk_X509_value(chain, chain_length-1);

    ca = std::string(X509_NAME_oneline(X509_get_issuer_name(datacert), buf, 1000));

    if ((!(X509_NAME_cmp(X509_get_subject_name(datacert), X509_get_issuer_name(datacert)))) && (chain_length >= 2))
      datacert = sk_X509_value(chain, chain_length - 2);

    subject = std::string(X509_NAME_oneline(X509_get_subject_name(datacert), buf, 1000));    
  }

  if (ca.empty() || subject.empty())
    return false;

  OBJ_create("1.3.6.1.4.1.8005.100.100.1","VOMS","Voms");
  OBJ_create("1.3.6.1.4.1.8005.100.100.2","INCFILE","IncFile");
  OBJ_create("1.3.6.1.4.1.8005.100.100.3","VO","Vo");

  nid  = OBJ_txt2nid("Voms");
  nidf = OBJ_txt2nid("IncFile");
  nidv = OBJ_txt2nid("Vo");

  index = X509_get_ext_by_NID(cert, nidf, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext){
      file = std::string((char *)(ext->value->data),ext->value->length);
      found = true;
    }
  }

  index = X509_get_ext_by_NID(cert, nidv, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext) {
      vo = std::string((char *)(ext->value->data),ext->value->length);
      found = true;
    }
  }

  index = X509_get_ext_by_NID(cert, nid, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext) {
      buffer = std::string((char *)(ext->value->data),ext->value->length);
      found = true;
    }
  }

#if 1
  int i =sk_X509_EXTENSION_num(cert->cert_info->extensions);
  X509_EXTENSION *ex;

  for (int j = 0; j < i; j++) {
    ex = sk_X509_EXTENSION_value(cert->cert_info->extensions, j);
    OBJ_obj2txt(buf, 1000, ex->object, 1);
#if 0
    if (strcmp(buf,"1.3.6.1.4.1.8005.100.100.3") == 0) {
      vo = std::string((char *)ex->value->data, ex->value->length);
      found = true;
    }
    if (strcmp(buf,"1.3.6.1.4.1.8005.100.100.2") == 0) {
      file = std::string((char *)ex->value->data, ex->value->length);
      found = true;
    }
#endif
    if (buffer.empty() && strcmp(buf,"1.3.6.1.4.1.8005.100.100.1") == 0) {
      buffer = std::string((char *)ex->value->data, ex->value->length);
      found = true;
    }
  }

#endif

  if (found)
    return true;

  if (how == RECURSE_CHAIN) {
    /*
     * May need to travel up the chain.
     */


      
    chain_length = sk_X509_num(chain);
    while (position < chain_length) {
      cert = sk_X509_value(chain,position);
      //      index = X509v3_get_ext_by_NID(cert->cert_info->extensions, nid, 0);
      {
	int j = 0;
	int i =sk_X509_EXTENSION_num(cert->cert_info->extensions);
	X509_EXTENSION *ex;
	char buf[1000];
	bool found = false;
	for (j = 0; j < i; j++) {
	  ex = sk_X509_EXTENSION_value(cert->cert_info->extensions, j);
	  OBJ_obj2txt(buf, 1000, ex->object, 1);
	  if (strcmp(buf,"1.3.6.1.4.1.8005.100.100.3") == 0) {
	    vo = std::string((char *)ex->value->data, ex->value->length);
	  }
	  if (strcmp(buf,"1.3.6.1.4.1.8005.100.100.2") == 0) {
	    file = std::string((char *)ex->value->data, ex->value->length);
	  }
	  if (strcmp(buf,"1.3.6.1.4.1.8005.100.100.1") == 0) {
	    buffer = std::string((char *)ex->value->data, ex->value->length);
	    found = true;
	  }
	}
	if (found)
	  return true;
      }
      /*
      if (index >= 0) {
	ext = X509_get_ext(cert,index);
	if (ext) {
	  buffer = std::string(ext->value->data, ext->value->length);
	  return true;
	}
      }
      */
      position++;
    } 
  }

  error = VERR_NOEXT;
  return false;
}




bool
verify(std::string message, struct vomsdata &voms, verror_type &error, std::string vdir, std::string cdir, std::string subject, std::string ca, bool verify)
{
  bool result = false;

  if (vdir.empty() || cdir.empty()) {
    error = VERR_NOINIT;
    return false;
  }

  ASN1_TIME *time1 = NULL, *time2 = NULL;

  do {
    struct ext_voms unformatted;

    error = VERR_PARSE;
    if (!unformat(message, unformatted, error))
      return false;

    if (!verify) {
      if (!parse_rest(unformatted.datalen, unformatted.real_data, 
                  unformatted.voms, error)) goto error;
      message=unformatted.extra_data;
      voms.data.push_back(unformatted.voms);
      continue;
    };

    error = VERR_VERIFY;
    if (!check(unformatted, error, vdir, cdir))
      return false;

#if 1
    /* Parse the date. */
    unsigned char *timebuf;

    timebuf = (unsigned char *)unformatted.voms.date1.c_str();
    time1 = d2i_ASN1_TIME(NULL, &timebuf, unformatted.voms.date1.size());

    timebuf = (unsigned char *)unformatted.voms.date2.c_str();
    time2 = d2i_ASN1_TIME(NULL, &timebuf, unformatted.voms.date2.size());

    /* Default error for following checks */
    error = VERR_TIME;
    if (!time1 || !time2)
      goto error;

    /* Verifies the date constraints */
    time_t ctime;
    time (&ctime);
    ctime += 300;

    if (((X509_cmp_current_time(time1) >=0) &&
	 (X509_cmp_time(time1, &ctime) >=  0)) || 
	(X509_cmp_current_time(time2) <= 0))
      goto error;
#endif

    /* Verifies that the cert identifies self */
    if (!subject.empty() && !ca.empty()) {
      error = VERR_IDCHECK;
      if (unformatted.voms.user != subject || unformatted.voms.userca != ca)
	goto error;
    }

    if (!parse_rest(unformatted.datalen, unformatted.real_data, unformatted.voms, error))
      goto error;

    error = VERR_EXTRAINFO;
    if (unformatted.voms.uri.empty() || unformatted.voms.voname.empty())
      goto error;

    message=unformatted.extra_data;
    voms.data.push_back(unformatted.voms);
  } while (!message.empty());

  result = true;
 error:
  if (time1)
    ASN1_TIME_free(time1);
  if (time2)
    ASN1_TIME_free(time2);
  return result;
}

static std::string getline(std::string &orig)
{
  std::string tmp;
  int pos = orig.find('\n');
  if (pos == std::string::npos)
    return "";

  tmp = orig.substr(0, pos);
  orig = orig.substr(pos+1);
  return tmp;
}

static bool
parse_rest(int len, std::string &data, struct voms &voms, verror_type &error)
{
  std::string to_parse = data.substr(0,len);
  std::string extra    = data.substr(len);

  data = extra;

  if (to_parse.find("NO DATA") == 0) {
    voms.type = TYPE_NODATA;
    return true;
  }

  if (to_parse.find("GROUP: ") == 0) {
    error = VERR_FORMAT;
    /* Standard data */
    voms.type = TYPE_STD;
    std::string group, role, cap;
    do {
      group = getline(to_parse);
      role = getline(to_parse);
      cap = getline(to_parse);
      if (!cap.empty()) {
	std::string g,r,c;
	int pos;

	if (group.find("GROUP: ") == 0)
	  g = group.substr(7);
	if (role.find("ROLE: ") == 0)
	  r = role.substr(6);
	if (cap.find("CAP: ") == 0)
	  c = cap.substr(5);
	if (g.empty() || r.empty() || c.empty())
	  return false;
	else {
	  struct data d = {g, r, c};
	  voms.std.push_back(d);
	}
      }
      else if (to_parse != "") 
	return false;
    } while (!cap.empty());
  }
  else {
    voms.type = TYPE_CUSTOM;
    voms.custom = to_parse;
  }

  return true;
}
      

/*
 * Function:
 *   unformat(data, len, key)
 *
 * Description:
 *   This function takes a buffer full of data returned by the format()
 *   (See) function, analyzes it for general correctness and rescinds it
 *   into its components, that are subsequently memorized into a collection
 *   structure.
 *
 * Parameters:
 *   'data' - Pointer to the buffer to analyze.
 *   'len'  - Length of the buffer.
 *   'key'  - Pointer to the server's public key.
 *
 * Result:
 *   Failure:
 *     NULL.
 *   Success:
 *     A pointer to a struct collection containing the broken down data
 *     that must be manually freed.  To do so, there is no need to also
 *     free its pointed data.
 */
static bool
unformat(const std::string data, struct ext_voms &results, verror_type &error)
{
  if (data.empty()) {
    error = VERR_NODATA;
    return false;
  }

  /*
   * Load the signature.
   */

  int sig_len = 0;
  bool ok;

  std::string is=data;

  ok  = parse(is, "SIGLEN:", sig_len);
  ok &= parse(is, "SIGNATURE:");

  error = VERR_PARSE;
  if (!ok)
    return false;

  results.voms.siglen = sig_len;
  results.voms.signature = is.substr(0,sig_len);
  is = results.signed_data = is.substr(sig_len);

  /*
   * Extracts data from the buffer.
   */

  ok  = parse(is, "USER:", results.voms.user);
  ok &= parse(is, "UCA:", results.voms.userca);
  ok &= parse(is, "SERVER:", results.voms.server);
  ok &= parse(is, "SCA:", results.voms.serverca);
  ok &= parse(is, "VO:", results.voms.voname);
  ok &= parse(is, "URI:", results.voms.uri);
  ok &= parse(is, "TIME1:" , results.voms.date1);
  ok &= parse(is, "TIME2:", results.voms.date2);
  ok &= parse(is, "DATALEN:", results.datalen);

  std::string::size_type pos = results.signed_data.find(is);

  if ((pos + results.datalen) > data.size())
    return false;
  else {
    results.real_data   = is.substr(0, results.datalen);
    results.extra_data  = is.substr(results.datalen);
    results.signed_data = results.signed_data.substr(0,pos+results.datalen);
  }
  return ok;
}

static bool
check(ext_voms &voms, verror_type &error,  std::string &vdir, std::string &cdir)
{
  DIR *dp;
  struct dirent *de;
  bool found  = false;
  bool result = false;

  BIO *in = NULL;
  X509 *x = NULL;

  error = VERR_DIR;
  if (!(dp = opendir(vdir.c_str())))
    return false;
 
  while (de = readdir(dp)) {
    char *name = de->d_name;

    if (name) {
      in = BIO_new(BIO_s_file());
      if (in) {
	std::string temp = vdir + "/" + name;
	if (BIO_read_filename(in, temp.c_str()) > 0) {
	  x = PEM_read_bio_X509(in, NULL, 0, NULL);
	  if (x) {
	    EVP_PKEY *key = X509_extract_key(x);
	    key = x->cert_info->key->pkey;
	    if (verify(key, voms.signed_data, voms.voms.signature)) {
	      found = true;
	      break;
	    }
	    else X509_free(x);
	  }
	}
      }
    }
  }

  error = VERR_SIGN;
  if (in) BIO_free(in);
  (void)closedir(dp);

  if (found) {
    char buf[2048];

    error = VERR_SERVER;
    /* check server subject */
    if (strcmp(X509_NAME_oneline(X509_get_subject_name(x), buf,2048),
	       voms.voms.server.c_str()) == 0) {
      /* check server ca */
      if (strcmp(X509_NAME_oneline(X509_get_issuer_name(x), buf,2048),
		 voms.voms.serverca.c_str()) == 0) {
	/* check cert */
	result = check_cert(x, error, cdir);
      }
    }
    X509_free(x);
  }

  return result;
}

static bool
check_cert(X509 *cert, verror_type &error,  std::string &cdir)
{
  X509_STORE *ctx = NULL;
  X509_STORE_CTX *csc = NULL;
  X509_LOOKUP *lookup = NULL;
  int i = 0;

  csc = X509_STORE_CTX_new();
  ctx = X509_STORE_new();
  error = VERR_MEM;
  if (ctx && csc) {
    X509_STORE_set_verify_cb_func(ctx,cb);
    ERR_load_crypto_strings();
#ifdef SIGPIPE
    signal(SIGPIPE,SIG_IGN);
#endif
    CRYPTO_malloc_init();
    SSLeay_add_all_algorithms();
    if (lookup = X509_STORE_add_lookup(ctx, X509_LOOKUP_file())) {
      X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT);
      if (lookup=X509_STORE_add_lookup(ctx,X509_LOOKUP_hash_dir())) {
	X509_LOOKUP_add_dir(lookup, cdir.c_str(), X509_FILETYPE_PEM);
	ERR_clear_error();
	error = VERR_VERIFY;
	X509_STORE_CTX_init(csc,ctx,cert,NULL);
	i = X509_verify_cert(csc);
      }
    }
  }
  if (ctx) X509_STORE_free(ctx);
  if (csc) X509_STORE_CTX_free(csc);

  if (i)
    return true;
  else
    return false;
}


static int MS_CALLBACK 
cb(int ok, X509_STORE_CTX *ctx)
{
  char buf[256];

  if (!ok) {
    X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),buf,256);
    if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ok=1;
    /* since we are just checking the certificates, it is
     * ok if they are self signed. But we should still warn
     * the user.
     */
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
    /* Continue after extension errors too */
    if (ctx->error == X509_V_ERR_INVALID_CA) ok=1;
    if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ok=1;
    if (ctx->error == X509_V_ERR_INVALID_PURPOSE) ok=1;
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
  }
  return(ok);
}


static bool 
id_data(std::string &subj, std::string &ca, verror_type &error)
{
  OM_uint32                   major_status = 0;
  OM_uint32                   minor_status = 0;

  gss_cred_id_t credential = GSS_C_NO_CREDENTIAL;

  /* acquire our credentials */
  major_status = globus_gss_assist_acquire_cred(&minor_status,
						GSS_C_BOTH,
						&credential);

  if(major_status != GSS_S_COMPLETE) {
    error = VERR_NOIDENT;
    return false;
  }

  char *tmp;

  tmp = getMCA(credential, globus(0));
  if (tmp)
    ca = std::string(tmp);

  tmp = get_globusid(credential);
  if (tmp)
    subj = std::string(tmp);

  return true;
}

static char *
get_globusid(gss_cred_id_t handle)
{
  char *globusid;
  char *globusid_tmp;
  gss_name_t server_name = GSS_C_NO_NAME;
  gss_buffer_desc server_buffer_desc = GSS_C_EMPTY_BUFFER;
  gss_buffer_t server_buffer = &server_buffer_desc;
  OM_uint32 major_status = 0;
  OM_uint32 minor_status = 0;
  OM_uint32 minor_status2 = 0;

  if ((major_status = gss_inquire_cred(&minor_status,
				       handle,
				       &server_name,
				       NULL, NULL, NULL)) == GSS_S_COMPLETE) {
    major_status = gss_display_name(&minor_status,
				    server_name, server_buffer, NULL);
    gss_release_name(&minor_status2, &server_name);
  }
  /*
   * The gssapi_cleartext does not implement gss_inquire_cred,
   * so fall back to using environment variable.
   */
  if (major_status == GSS_S_COMPLETE) {
    globusid = (char *)server_buffer_desc.value;
  }
  else {
    return NULL;
  }
  globusid_tmp = strdup(globusid);

  if (server_buffer_desc.value) {
    gss_release_buffer(&minor_status2, server_buffer);
  }
  return globusid_tmp;
}

