/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#include "config.h"

#define SUBPACKAGE "voms"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>

#include <openssl/evp.h>

#include "Server.h"

#include "gssapi.h"

#include "VOMSServer.h"

#include "options.h"
#include "data.h"
#include "pass.h"
#include "format.h"

#include "access_db_sql.h"

#include <map>
#include <string>
#include <algorithm>
#include <iostream>

#include "attribute.h"

const int DEFAULT_PORT = 15000;

typedef std::map<attrib, int> ordermap;

static ordermap ordering;

static std::string firstgroup="";
static std::string firstrole="";

static void
sigchld_handler(int sig)
{
  int save_errno = errno;
  pid_t pid;
  int status;

  while ((pid = waitpid(-1, &status, WNOHANG)) > 0 ||
	 (pid < 0 && errno == EINTR))
    ;

  signal(SIGCHLD, sigchld_handler);
  errno = save_errno;
}


static bool compare(const attrib &lhs, const attrib &rhs)
{
  ordermap::iterator lhi=ordering.find(lhs);
  ordermap::iterator rhi=ordering.find(rhs);

  if (lhi == ordering.end())
    return false;

  if (rhi == ordering.end())
    return true;

  return (lhi->second < rhi->second);
}

static void orderattribs(std::vector<attrib> &v)
{
  std::partial_sort(v.begin(), v.begin()+ordering.size(), v.end(), compare);
}

static bool tokenize(const std::string &message, int &validity, ordermap &ordering, std::string &command)
{
  std::string::size_type valid_end,
                         order_end;
  
  valid_end = message.find_first_not_of("0123456789");
  order_end = message.find_last_of(',');

  /* Validity and command are an obligatory parameter */
  if ((valid_end == std::string::npos) || (order_end == std::string::npos))
    return false;

  /* Discover validity */
  int position = 0;
  validity = 0;

  while (position < valid_end) {
    validity *= 10;
    validity += message.at(position++) - '0';
  }

  int order = 0;

  while (position != order_end) {
    position++;
    /* There is a specified ordering */
    std::string::size_type end_token = message.find_first_of(',',position);
    std::string attribute = message.substr(position, end_token - position);
    std::string::size_type divider = attribute.find(':');
    attrib temp;
    if (divider == std::string::npos) {
      if (firstgroup.empty()) {
	firstgroup = attribute;
	firstrole = "NULL";
      }
      temp.group = attribute;
      temp.role = "NULL";
    }
    else {
      if (firstgroup.empty()) {
	firstgroup = attribute.substr(0, divider);
	firstrole = attribute.substr(divider+1);
      }
      temp.group = attribute.substr(0, divider);
      temp.role  = attribute.substr(divider+1);
    }
    temp.cap = "";
    ordering.insert(std::make_pair<attrib,int>(temp,order));
    order++;
    position = end_token;
  }

  command = message.substr(position+1);
  return true;
}

VOMSServer::VOMSServer(int argc, char *argv[]) : sock(0,0,50,false),
						 logger(), validity(86400),
						 logfile("/var/log/voms"),
						 gatekeeper_test(false),
						 daemon_port(DEFAULT_PORT),
						 foreground(false),
						 globuspwd(""), globusid(""),
						 x509_cert_dir(""),
						 x509_cert_file(""),
						 x509_user_proxy(""),
						 x509_user_cert(""),
						 x509_user_key(""),
						 desired_name_char(""),
						 username("voms"),
						 dbname("voms"), passfile(""),
						 voname("unspecified"),
						 uri(""),
						 version(0),
						 subject(""),
						 ca("")
{
  struct stat statbuf;

  signal(SIGCHLD, sigchld_handler);

  if ((stat("/etc/nologin", &statbuf)) == 0) {
    throw VOMSInitException();
  }

  std::string fakeuri = "";
  bool progversion = false;
  int blog = 50;

  struct option opts[] = {
    {"test",            0, (int *)&gatekeeper_test,   OPT_BOOL},
    {"conf",            1, NULL,                      OPT_CONFIG},
    {"port",            1, &daemon_port,              OPT_NUM},
    {"logfile",         1, (int *)&logfile,           OPT_STRING},
    {"globusid",        1, (int *)&globusid,          OPT_STRING},
    {"globuspwd",       1, (int *)&globuspwd,         OPT_STRING},
    {"x509_cert_dir",   1, (int *)&x509_cert_dir,     OPT_STRING},
    {"x509_cert_file",  1, (int *)&x509_cert_file,    OPT_STRING},
    {"x509_user_proxy", 1, (int *)&x509_user_proxy,   OPT_STRING},
    {"x509_user_cert",  1, (int *)&x509_user_cert,    OPT_STRING},
    {"x509_user_key",   1, (int *)&x509_user_key,     OPT_STRING},
    {"desired_name",    1, (int *)&desired_name_char, OPT_STRING},
    {"foreground",      0, (int *)&foreground,        OPT_BOOL},
    {"username",        1, (int *)&username,          OPT_STRING},
    {"timeout",         1, &validity,                 OPT_NUM},
    {"dbname",          1, (int *)&dbname,            OPT_STRING},
    {"passfile",        1, (int *)&passfile,          OPT_STRING},
    {"vo",              1, (int *)&voname,            OPT_STRING},
    {"uri",             1, (int *)&fakeuri,           OPT_STRING},
    {"globus",          1, &version,                  OPT_NUM},
    {"version",         0, (int *)&progversion,       OPT_BOOL},
    {"backlog",         1, &blog,                     OPT_NUM},
    {0, 0, 0, 0}
  };

  /*
   * Parse the command line arguments
   */
  
  set_usage("[-conf parmfile] [-foreground] [-port port]\n"
	    "[-logfile file] [-passfile file] [-vo voname]\n"
	    "[-globusid globusid] [-globuspwd file] [-globus version]\n"
	    "[-x509_cert_dir path] [-x509_cert_file file]\n"
	    "[-x509_user_cert file] [-x509_user_key file]\n"
	    "[-dbname name] [-username name] [-timeout limit]\n"
	    "[-x509_user_proxy file] [-test] [-uri uri]\n");

  if (!getopts(argc, argv, opts))
    throw VOMSInitException();

  if (progversion) {
    std::cout << SUBPACKAGE << "\nVersion: " << VERSION << std::endl;
    std::cout << "Compiled: " << __DATE__ << " " << __TIME__ << std::endl;
    exit(0);
  }

  if (!getpasswd(passfile))
    throw VOMSInitException();


  version = globus(version);
  if (version == 0) {
    std::cerr << "Unable to discover Globus Version: Trying for 2.2" 
	      << std::endl << std::flush;
    version = 22;
  }

  if (fakeuri == "") {
    int   ok;

    int   hostnamesize = 50;
    char *hostname = new char[1];
    do {
      delete[] hostname;
      hostname = new char[hostnamesize];
      ok = gethostname(hostname, hostnamesize);
      hostnamesize += 50;
    } while (ok);
    std::string temp;

    uri = std::string(hostname) + ":" + stringify(daemon_port, temp);
    delete[] hostname;
  }
  else
    uri = fakeuri;

  sock = GSISocketServer(daemon_port, version, blog);

  setenv("GLOBUSID", globusid.c_str(), 1);

  /*
   * Dont use default env proxy cert for gatekeeper if run as root
   * this might get left over. You can still use -x509_user_proxy
   */

  unsetenv("X509_USER_PROXY");

  if (!globuspwd.empty())
  {
    setenv("GLOBUSPWD", globuspwd.c_str(), 1);
  }

  if (!x509_cert_dir.empty())
  {
    setenv("X509_CERT_DIR", x509_cert_dir.c_str(), 1);
  }
  if (!x509_cert_file.empty())
  {
    setenv("X509_CERT_FILE", x509_cert_file.c_str(), 1);
  }
  if (!x509_user_proxy.empty())
  {
    setenv("X509_USER_PROXY", x509_user_proxy.c_str(), 1);
  }

  if (!x509_user_cert.empty())
  {
    setenv("X509_USER_CERT", x509_user_cert.c_str(), 1);
  }
  if (!x509_user_key.empty())
  {
    setenv("X509_USER_KEY", x509_user_key.c_str(), 1);
  }
#if (__GNUC__ >=3)
  int fd = open(logfile.c_str(), O_WRONLY | O_APPEND | O_CREAT);
  if (fd != -1)
    logger.open(fd);
  else
  {
    std::cerr << "Logging startup failure" << std::endl;
    throw VOMSInitException();
  }
#else
  logger.open(logfile.c_str(), std::ios::out | std::ios::app);
  if (!logger)
  {
    std::cerr << "Logging startup failure" << std::endl;
    throw VOMSInitException();
  }

  /* attaches stderr to the log */
  int fd=logger.rdbuf()->fd();
#endif

  FILE *f=fdopen(fd, "a");
  if (!f) {
    std::cerr << strerror(errno) << std::endl;
    std::cerr << "GSI Logging setup failure" << std::endl;
    throw VOMSInitException();
  }

  sock.RedirectGSIOutput(f);
  sock.SetFlags(GSS_C_MUTUAL_FLAG | GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG);
  logger << argv[0] << " pid=" << getpid()
	 << " starting at " << timestamp() << std::endl;
  logger << "URI: " << uri << std::endl;
  logger << "Detected Globus Version: " << version << std::endl;
}

VOMSServer::~VOMSServer() {}

void VOMSServer::Run(void)
{
  int status;

  std::cerr << std::flush;


  if (daemon(0,0))
    exit(0);

  try {
    logger << "try to open" << std::endl;
    sock.Open();

    logger << "Started: " << sock.own_subject << std::endl 
	   << "CA: " << sock.own_ca << std::endl;

    subject = sock.own_subject;
    ca = sock.own_ca;

    for (;;) {
      if (sock.Listen()) {

	std::string user    = sock.peer_subject;
	std::string userca  = sock.peer_ca;

	logger << "At " << timestamp()
	       << " Received contact from: " << std::endl
	       << " user: " << user << std::endl
	       << " ca: " << userca << std::endl;
	  
	if (!gatekeeper_test) {
	  if (!fork()) {
	    sock.CloseListener();
	    logger << "voms executor" << std::endl;
	    bool r = Execute(user, userca, sock.own_key);
	    sock.Close();
	    exit(0);
	  }
	  else {
	    logger << "voms receiver" << std::endl;
	    sock.CloseListened();
	  }
	}
	else {
	  Execute(user, userca, sock.own_key);
	  sock.CloseListened();
	}
      }
    }
  }
  catch (...) {}
}

bool  
VOMSServer::Execute(const std::string &client_name, const std::string &ca_name, EVP_PKEY *key)
{
    std::string message;

    if (!sock.Receive(message)) {
      logger << "Unable to decode request." << std::endl;
      return false;
    }

    logger << "Received Request: " << message << std::endl;

    std::string command;
    int requested = 0;

    if (!tokenize(message, requested, ordering, command)) {
      logger << "Unable to interpret command: " << message << std::endl;
      return false;
    }

    const char *comm = command.c_str();
    std::vector<attrib> res;
    std::string data = "";
    std::string header = "";
    std::string tmp="";
    bool result = false;

    /* Interpret user requests */

    if (requested != 0)
      validity = (validity < requested ? validity : requested);

    if (validity < requested)
      header = "1SERVER: "+uri+"\nValidity shortened to " + stringify(validity, tmp) + " seconds!\n\1";

    logger << "New command: " << command << std::endl;
      
    /* Interpret request by first character */
    switch (*comm) {
    case 'A':
      result =
	get_all(client_name, ca_name, dbname, username, passwd(),res);
      break;

    case 'R':
      result =
	get_role(client_name, ca_name, comm+1, dbname, username, passwd(),res);
      break;
	
    case 'G':
      result=
	get_group(client_name, ca_name, comm+1, dbname, username, passwd(),res);
      break;

    case 'B':
      result =
	get_group_and_role(client_name, ca_name, comm+1, dbname, username, passwd(),res);
      break;

    case 'S':
      result =
	special(client_name, ca_name, comm+1, dbname, username, passwd(),data);
      break;

    case 'L':
      result =
	listspecial(dbname, username, passwd(), data);
      break;

    default:
      result = false;
      logger << "Unknown Command!" << std::endl;
      break;
    }

    if (result && !res.empty()) {
      orderattribs(res);
    }

    std::string falsefirst="";
    if (!firstgroup.empty()) {
      std::vector<attrib>::iterator j = res.begin();
      if (j != res.end()) {
	if ((j->group != firstgroup) || (j->role != firstrole))
	  falsefirst = std::string("0GROUP: ") + j->group + "\nROLE: " + j->role + "\n\1";
      }
      if (j == res.end() && (!firstgroup.empty() || !firstrole.empty()))
	falsefirst = std::string("0NO GROUP/ROLE FOUND!\n\1");
    }

    for (std::vector<attrib>::iterator i = res.begin(); i != res.end(); i++)
      data += "GROUP: " + i->group + "\nROLE: " + i->role + "\nCAP: " + i->cap + "\n";

    if (!result) {
      logger << "Error in executing request!" << std::endl;
      sock.Send("");
      return false;
    }

    if (!data.empty()) {
      std::string formatted = "";
      result = format(data, client_name, ca_name, subject, ca, key, 
		      validity, voname, uri, formatted);

      if (!result) {
	sock.Send("");
	return false;
      }

      if (!falsefirst.empty())
	formatted = falsefirst + formatted;

      formatted = header + formatted;

      logger << "Request Result:" << std::endl
	     << data << std::endl << std::flush;
      logger << "Sent DATA:" << std::endl
	     << formatted << std::endl << std::flush;
      sock.Send(formatted);
    }
    else
      sock.Send("2You are not a member of the " + voname + " VO!\n\1");
    return true;
}
